import h5py
import numpy
import os
import pytest
from tempfile import TemporaryDirectory

@pytest.fixture(scope='module')
def simple_h5_file():
    with TemporaryDirectory() as td:
        f = h5py.File(os.path.join(td, 'example.h5'), 'w')
        f.create_group('group1')
        f.create_dataset('group1/subgroup1/dataset1', (200,), dtype='u8')
        f.create_dataset('group1/subgroup1/dataset2', (2, 128, 500), dtype='f4')
        f.create_dataset('group1/subgroup2/dataset1', (12,), dtype='i2')
        f.create_dataset('group1/scalar', shape=(), dtype='i4')
        f.create_dataset('group1/empty', shape=None, dtype='i4')
        f.create_dataset('compound',
            data=numpy.array([(1, 0.5), (3, 0.090)],
                             dtype=numpy.dtype([('count', 'u8'), ('amount', 'f4')])))
        f.create_dataset('prose',
            data=numpy.array(['Lorem ipsum dolor sit amet', 'consectetur adipiscing elit'],
                             dtype=h5py.special_dtype(vlen=str)))
        f['group1'].attrs['string'] = 'foo'
        f['group1'].attrs['array'] = numpy.zeros((3, 4))
        f['synonyms/folder'] = f['group1/subgroup1']
        f['synonyms/values'] = f['group1/subgroup1/dataset1']
        yield f
        f.close()

@pytest.fixture()
def file_with_custom_float(tmp_path):
    # create a custom type with larger bias
    mytype = h5py.h5t.IEEE_F16LE.copy()
    mytype.set_fields(14, 9, 5, 0, 9)
    mytype.set_size(2)
    mytype.set_ebias(53)
    mytype.lock()

    shape = (3, 4)
    space = h5py.h5s.create_simple(shape)

    with h5py.File(str(tmp_path / 'sample.h5'), 'w') as f:
        h5py.h5d.create(f.id, b'a', mytype, space)
        yield f



"""Shell integration to tab complete paths inside an HDF5 file"""
import os
import shutil
import subprocess

pjoin = os.path.join
pkgdir = os.path.dirname(os.path.abspath(__file__))

def install_hooks():
    """Install hooks for bash & zsh to complete paths in files"""
    data_home = os.environ.get('XDG_DATA_HOME', '') \
                or os.path.expanduser('~/.local/share')

    # Bash
    if shutil.which('bash'):
        src = pjoin(pkgdir, 'completion.bash')
        dst_dir = pjoin(data_home, 'bash-completion/completions')
        os.makedirs(dst_dir, exist_ok=True)
        dst = pjoin(dst_dir, 'h5glance')
        shutil.copy(src, dst)
        print("Copied", dst)

    # Zsh
    if shutil.which('zsh'):
        src = pjoin(pkgdir, 'completion.zsh')
        dst_dir = pjoin(data_home, 'zsh-completions')
        os.makedirs(dst_dir, exist_ok=True)
        dst = pjoin(dst_dir, '_h5glance')
        shutil.copy(src, dst)
        print("Copied", dst)

        stdout = subprocess.check_output(['zsh', '-i', '-c', 'echo $FPATH'])
        if dst_dir not in stdout.decode('utf-8').split(':'):
            with open(os.path.expanduser('~/.zshrc'), 'a') as f:
                f.write('\nfpath=("{}" $fpath)\ncompinit\n'.format(dst_dir))
            print("Added {} to fpath in ~/.zshrc".format(dst_dir))


if __name__ == '__main__':
    install_hooks()



import io
import os
import re
import pytest

from h5glance import terminal

@pytest.fixture(scope='module', autouse=True)
def no_color():
    os.environ['H5GLANCE_COLORS'] = '0'

def test_tree_view(simple_h5_file):
    sio = io.StringIO()
    tvb = terminal.TreeViewBuilder()
    tree = tvb.object_node(simple_h5_file, simple_h5_file.filename)
    terminal.print_tree(tree, file=sio)
    out = sio.getvalue()
    assert '└dataset2' in out
    assert 'uint64' in out
    assert '128 × 500' in out

def test_attributes(simple_h5_file):
    tvb = terminal.TreeViewBuilder()
    r1, c1 = tvb.object_node(simple_h5_file['group1'], 'group1')
    assert '2 attributes' in r1
    assert 'attributes' not in c1[0][0]

    tvb2 = terminal.TreeViewBuilder(expand_attrs=True)
    r2, c2 = tvb2.object_node(simple_h5_file['group1'], 'group1')
    assert 'attributes' not in r2
    assert '2 attributes' in c2[0][0]
    attr_lines = {node[0] for node in c2[0][1]}
    assert len(attr_lines) == 2

def test_dataset_info(simple_h5_file):
    sio = io.StringIO()
    terminal.print_dataset_info(simple_h5_file["/group1/subgroup1/dataset2"], file=sio)
    out = sio.getvalue()
    assert 'dtype: float32' in out
    assert 'shape: 2 × 128 × 500' in out

def test_dataset_slice(simple_h5_file):
    sio = io.StringIO()
    terminal.print_dataset_info(simple_h5_file["/group1/subgroup1/dataset2"],
                                slice_expr='0,0,5:9', file=sio)
    out = sio.getvalue()
    assert '[0,0,5:9]' in out
    assert '[0. 0. 0. 0.]' in out

def test_compound_types(simple_h5_file):
    sio = io.StringIO()
    terminal.print_dataset_info(simple_h5_file["compound"], file=sio)
    out = sio.getvalue()
    assert 'dtype: (count: uint64, amount: float32)' in out
    assert re.search(r'\(\s*3\s*,\s*0.09\s*\)', out)

def test_custom_float(file_with_custom_float):
    sio = io.StringIO()
    terminal.print_dataset_info(file_with_custom_float["a"], file=sio)
    out = sio.getvalue()
    assert 'dtype: custom 2-byte float' in out

def test_vlen_types(simple_h5_file):
    sio = io.StringIO()
    terminal.print_dataset_info(simple_h5_file["prose"], file=sio)
    out = sio.getvalue()
    assert 'dtype: UTF-8 str' in out
    assert "[b'Lorem ipsum dolor sit amet' b'consectetur adipiscing elit']" in out

def test_hard_links(simple_h5_file):
    sio = io.StringIO()
    tvb = terminal.TreeViewBuilder()
    tree = tvb.object_node(simple_h5_file, simple_h5_file.filename)
    terminal.print_tree(tree, file=sio)
    out = sio.getvalue()
    assert 'folder\t= /group1/subgroup1' in out
    assert 'values\t= /group1/subgroup1/dataset1' in out

def test_max_depth(simple_h5_file):
    sio = io.StringIO()
    tvb = terminal.TreeViewBuilder()
    tree = tvb.object_node(simple_h5_file, simple_h5_file.filename, max_depth=2)
    terminal.print_tree(tree, file=sio)
    out = sio.getvalue()
    assert 'prose\t[UTF-8 string: 2]' in out  # depth=0
    assert 'scalar\t[int32: scalar]' in out  # depth=1
    assert 'subgroup1\t(2 children)' in out  # depth=2
    assert 'dataset1\t[uint64: 200]' not in out  # depth=3

def test_completer(simple_h5_file):
    comp = terminal.H5Completer(simple_h5_file)
    # Complete groups
    assert set(comp.completions('group1/sub')) == {'group1/subgroup1/', 'group1/subgroup2/'}
    # Complete a dataset - no trailing slash
    assert set(comp.completions('group1/subgroup2/')) == {'group1/subgroup2/dataset1'}
    # Case insensitive completion
    assert set(comp.completions('GRO')) == {'group1/'}



from functools import lru_cache
from h5py import h5t
import numpy as np
from .utils import fmt_shape

cset_names = {h5t.CSET_ASCII: 'ASCII', h5t.CSET_UTF8: 'UTF-8'}

# h5py TypeIDs are not consistently hashable, so we can't use them as dict keys.
# Instead, we look up by size, and then check equality with the standard types.
@lru_cache()
def int_types_by_size():
    d = {}
    for tc in np.typecodes['AllInteger']:
        _add_typecode(tc, d)
    return d

@lru_cache()
def float_types_by_size():
    d = {}
    for tc in np.typecodes['Float']:
        _add_typecode(tc, d)
    return d

def _add_typecode(tc, sizes_dict):
    dt_le = np.dtype('<' + tc)
    dt_be = np.dtype('>' + tc)

    entries = sizes_dict.setdefault(dt_le.itemsize, [])
    entries.append((h5t.py_create(dt_le), dt_le.name))
    entries.append((h5t.py_create(dt_be), dt_be.name + ' (big-endian)'))


def fmt_dtype(hdf_dt):
    """Get a (preferably short) string describing an HDF5 datatype"""
    size = hdf_dt.get_size()

    if isinstance(hdf_dt, h5t.TypeIntegerID):
        # Check normal int & uint dtypes first
        for candidate, descr in int_types_by_size().get(size, ()):
            if hdf_dt == candidate:
                return descr
        un = 'un' if hdf_dt.get_sign() == h5t.SGN_NONE else ''
        return "{}-byte {}signed integer".format(size, un)
    elif isinstance(hdf_dt, h5t.TypeFloatID):
        # Check normal float dtypes first
        for candidate, descr in float_types_by_size().get(size, ()):
            if hdf_dt == candidate:
                return descr
        return "custom {}-byte float".format(size)
    elif isinstance(hdf_dt, h5t.TypeBitfieldID):
        return "{}-byte bitfield".format(size)
    elif isinstance(hdf_dt, h5t.TypeTimeID):
        return "time"  # NB. time datatype is deprecated
    elif isinstance(hdf_dt, h5t.TypeOpaqueID):
        s = "{}-byte opaque".format(size)
        tag = hdf_dt.get_tag()
        if tag:
            s += ' ({})'.format(tag.decode('utf-8', 'replace'))
        return s
    elif isinstance(hdf_dt, h5t.TypeStringID):
        cset = cset_names.get(hdf_dt.get_cset(), '?cset')
        if hdf_dt.is_variable_str():
            return "{} string".format(cset)
        else:
            return "{}-byte {} string".format(size, cset)
    elif isinstance(hdf_dt, h5t.TypeVlenID):
        return "vlen array of " + fmt_dtype(hdf_dt.get_super())
    elif isinstance(hdf_dt, h5t.TypeArrayID):
        shape = fmt_shape(hdf_dt.get_array_dims())
        return "{} array of {}".format(shape, fmt_dtype(hdf_dt.get_super()))

    elif isinstance(hdf_dt, h5t.TypeCompoundID):
        return "({})".format(", ".join(
            "{}: {}".format(
                hdf_dt.get_member_name(i).decode('utf-8', 'replace'),
                fmt_dtype(hdf_dt.get_member_type(i))
            )
            for i in range(hdf_dt.get_nmembers())
        ))
    elif isinstance(hdf_dt, h5t.TypeEnumID):
        nmembers = hdf_dt.get_nmembers()
        if nmembers >= 5:
            return "enum ({} options)".format(nmembers)
        else:
            return "enum ({})".format(", ".join(
                hdf_dt.get_member_name(i).decode('utf-8', 'replace')
                for i in range(nmembers)
            ))
    elif isinstance(hdf_dt, h5t.TypeReferenceID):
        return "region ref" if hdf_dt == h5t.STD_REF_DSETREG else "object ref"

    return "unrecognised {}-byte datatype".format(size)


def dtype_description(hdf_dt):
    """A slightly longer description, suitable for a tooltip

    Can return None
    """
    size = hdf_dt.get_size()

    if isinstance(hdf_dt, h5t.TypeIntegerID):
        # Check normal int & uint dtypes first
        for candidate, descr in int_types_by_size().get(size, ()):
            if hdf_dt == candidate:
                un = 'un' if hdf_dt.get_sign() == h5t.SGN_NONE else ''
                return '{}-bit {}signed integer'.format(size * 8, un)

    elif isinstance(hdf_dt, h5t.TypeFloatID):
        # Check normal float dtypes first
        for candidate, descr in float_types_by_size().get(size, ()):
            if hdf_dt == candidate:
                return '{}-bit floating point'.format(size * 8)

    return None



"""Helper functions
"""
import h5py

_mapping = {
    h5py.File: "file",
    h5py.Group: "group",
    h5py.Dataset: "dataset",
}


def register_h5pyd():
    import h5pyd
    _mapping[h5pyd.File] = "file"
    _mapping[h5pyd.Group] = "group"
    _mapping[h5pyd.Dataset] = "dataset"


def check_class(obj_class):
    """Check registered class"""
    lib = obj_class.__module__.split(".")[0]
    if lib == "h5pyd":
        register_h5pyd()
    else:
        _mapping[obj_class] = None


def get_h5py_kind(obj):
    """Returns the h5py-like kind of an object.

    The result can be `file`, `group`, `dataset` or `None` if the object is not
    an h5py-like object.
    """
    obj_class = type(obj)
    if obj_class not in _mapping:
        check_class(obj_class)
    return _mapping.get(obj_class, None)


def is_file(obj):
    """Returns true if the object is a h5py-like file."""
    kind = get_h5py_kind(obj)
    return kind == "file"


def is_dataset(obj):
    """Returns true if the object is a h5py-like dataset."""
    kind = get_h5py_kind(obj)
    return kind == "dataset"


def is_group(obj):
    """Returns true if the object is a h5py-like group."""
    kind = get_h5py_kind(obj)
    return kind in ["file", "group"]


def fmt_shape(shape):
    if shape is None:
        return "empty"
    if shape == ():
        return "scalar"
    return " × ".join(('Unlimited' if n is None else str(n)) for n in shape)



from h5glance import datatypes
from h5py import h5t
import numpy as np

def test_standard_float():
    ft = h5t.py_create(np.dtype('<f4'))
    assert datatypes.fmt_dtype(ft) == 'float32'
    assert datatypes.dtype_description(ft) == '32-bit floating point'

def test_custom_float():
    f1 = h5t.IEEE_F16LE.copy()
    f1.set_fields(14, 9, 5, 0, 9)
    f1.set_size(2)
    f1.set_ebias(53)
    assert datatypes.fmt_dtype(f1) == 'custom 2-byte float'

def test_standard_int():
    it = h5t.py_create(np.dtype('<i4'))
    assert datatypes.fmt_dtype(it) == 'int32'
    assert datatypes.dtype_description(it) == '32-bit signed integer'

    ut = h5t.py_create(np.dtype('>u8'))
    assert datatypes.fmt_dtype(ut) == 'uint64 (big-endian)'
    assert datatypes.dtype_description(ut) == '64-bit unsigned integer'

def test_custom_int():
    i1 = h5t.STD_U32LE.copy()
    i1.set_size(6)
    assert datatypes.fmt_dtype(i1) == '6-byte unsigned integer'

def test_string():
    # vlen string
    vst = h5t.py_create(h5t.string_dtype(encoding='utf-8'), logical=True)
    assert datatypes.fmt_dtype(vst) == 'UTF-8 string'

    # fixed-length string
    fst = h5t.py_create(h5t.string_dtype(encoding='ascii', length=3))
    assert datatypes.fmt_dtype(fst) == '3-byte ASCII string'

def test_compound():
    ct_n = np.dtype([('x', np.float32), ('y', np.float32)])
    ct_h = h5t.py_create(ct_n)
    assert datatypes.fmt_dtype(ct_h) == '(x: float32, y: float32)'

def test_enum():
    et_n = h5t.enum_dtype({'apple': 1, 'banana': 2})
    et_h = h5t.py_create(et_n, logical=True)
    assert datatypes.fmt_dtype(et_h) == 'enum (apple, banana)'

def test_vlen():
    vt_n = h5t.vlen_dtype(np.dtype('<i2'))
    vt_h = h5t.py_create(vt_n, logical=True)
    assert datatypes.fmt_dtype(vt_h) == 'vlen array of int16'

def test_array():
    at_n = np.dtype((np.float64, (3, 4)))
    at_h = h5t.py_create(at_n)
    assert datatypes.fmt_dtype(at_h) == '3 × 4 array of float64'



import h5py
import os

from .html import h5obj_to_html
from .terminal import group_to_str

class H5Glance:
    """View an HDF5 object in a Jupyter notebook"""
    def __init__(self, obj):
        self.obj = obj

    def _repr_html_(self):
        return h5obj_to_html(self.obj)

    def __repr__(self):
        if isinstance(self.obj, (str, bytes, os.PathLike)):
            with h5py.File(self.obj, 'r') as f:
                return repr(H5Glance(f))
        return group_to_str(self.obj, max_depth=1)

def install_ipython_h5py_display(html=True, text=True):
    """Call inside IPython to install HTML/text views for h5py groups and files"""
    from IPython import get_ipython
    ip = get_ipython()
    if ip is None:
        raise EnvironmentError("This function is to be called in IPython")

    if html:
        html_formatter = ip.display_formatter.formatters['text/html']
        html_formatter.for_type(h5py.Group, h5obj_to_html)
    if text:
        text_formatter = ip.display_formatter.formatters['text/plain']
        text_formatter.for_type(h5py.Group, pretty_print_group)
        text_formatter.for_type(h5py.File, pretty_print_group)

def pretty_print_group(obj, p, cycle):
    p.text(group_to_str(obj))
